
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>



#include "la/la_interface.h"
#include "lxt/lxt_write.h"
#include "sampler.h"
#include "xml.h"

/* Maximum von a und b bestimmen */
#define max(a, b) ((a < b) ? (b) : (a))

/*
char *sampleRates[] = {
	"600 MSamples/s (1,66 ns)",		500 000 000
	"150 MSamples/s (6,66 ns)",
	"50 MSamples/s (20 ns)",
	"25 MSamples/s (40 ns)",
	"10 MSamples/s (100 ns)",		 10 000 000
	"5 MSamples/s (200 ns)",
	"2,5 MSamples/s (400 ns)",
	"1 MSamples/s (1 us)",			  1 000 000
	"500 KSamples/s (2 us)",
	"250 KSamples/s (4 us)",
	"100 KSamples/s (10 us)",		    100 000
	"50 KSamples/s (20 us)",
	"25 KSamples/s (40 us)",
	"10 KSamples/s (100 us)",		     10 000
	"5 KSamples/s (200 us)",
	"2,5 KSamples/s (400 us)",
	"1 KSamples/s (1 ms)",			      1 000
	"500 Samples/s (2 ms)",
	"250 Samples/s (4 ms)",
	"100 Samples/s (10 ms)",			100
};
*/
 
/*
Integer Wert aus XML Datei lesen.
Im Fehlerfall eine Meldung ausgeben und einen Default-Wert annehmen.
*/
unsigned long read_long(const char *name, const char *xpath, unsigned long default_value)
{
	char buf[100];

	/* Versuche Wert zu ermitteln */
	if (xml_get_string(xpath, buf, 100) != 0)
	{
		/* Benutze Standardwert bei Fehler */
		printf("Warnung: Wert für %s konnte nicht gelesen werden!\n", name);
		return default_value;
	}
	else
	{
		/* Ergenis in Zahl umwandeln */
		return atol(buf);
	}
}

/*
Signal aus Konfigurationsdatei lesen.
*/
void read_la_signal(struct la_signal *signal, int i)
{
	int j;
	char xpath[100], buf[100];

	/* Signalname einlesen */
	sprintf(xpath, "//la/signals/signal[%d]/@name", i + 1);
	xml_get_string(xpath, buf, 100);

	signal->name = (char *)malloc(strlen(buf) + 1);
	strcpy(signal->name, buf);

	/* Channels einlesen */
	sprintf(xpath, "//la/signals/signal[%d]/ch", i + 1);
	signal->num_channels = max(xml_get_element_count(xpath), 0);
	signal->channels = (unsigned char *)malloc(signal->num_channels * sizeof(unsigned char));

	for (j = 0; j < signal->num_channels; j++)
	{
		sprintf(xpath, "//la/signals/signal[%d]/ch[%d]/text()", i + 1, j + 1);
		signal->channels[j] = (unsigned char) read_long("Channel", xpath, 0);
	}
}

/* Condition Attribut eines Trigger Channel auswerten und
   entsprechende Werte in der Trigger Struktur setzen */
int parse_la_trigger_condition(char *condition, struct la_trigger *trigger)
{
	char *token;

	trigger->low 		= 0;
	trigger->high 		= 0;
	trigger->falling_edge	= 0;
	trigger->rising_edge 	= 0;

	/* Condition Attribut an Leerzeichen in Teilstrings teilen */
	token = strtok(condition, " ");
	/* Für jeden Teilstring... */
	while (token != NULL)
	{
		/* Auslösen wenn Signal auf high ist */
		if (strcmp(token, "high") == 0)
		{
			trigger->high = 1;
		}
		/* Auslösen wenn Signal auf low ist */
		if (strcmp(token, "low") == 0)
		{
			trigger->low = 1;
		}
		/* Auslösen bei steigender Flanke */
		if (strcmp(token, "rising_edge") == 0)
		{
			trigger->rising_edge = 1;
		}
		/* Auslösen bei fallende Flanke */
		if (strcmp(token, "falling_edge") == 0)
		{
			trigger->falling_edge = 1;
		}

		token = strtok(NULL, " ");
	}

	/* Werte auf Gültigkeit prüfen */
	/* High/Low bzw. RE/FE sind jeweils kombinierbar, aber High/Low und RE/FE darf nicht gleichzeitig gesetzt sein */
	if (!((trigger->low | trigger->high) ^ (trigger->falling_edge | trigger->rising_edge)))
	{
		trigger->low		= 0;
		trigger->high		= 0;
		trigger->falling_edge	= 0;
		trigger->rising_edge	= 0;

		return -1;
	}

	return 0;
}

/* Trigger aus Konfigurationsdatei lesen */
int read_la_trigger(struct la_trigger *trigger, char A_or_B)
{
	int i;
	int count, channel;
	char xpath[100], condition[100];

	sprintf(xpath, "//la/trigger/trigger%c/ch", A_or_B);
	count = xml_get_element_count(xpath);

	for (i = 0; i < count; i++)
	{
		/* Condition Attribut lesen */
		sprintf(xpath, "//la/trigger/trigger%c/ch[%d]/@when", A_or_B, i + 1);
		if (xml_get_string(xpath, condition, 100) != 0)
		{
			continue;
		}

		/* Channel Nummer lesen */
		sprintf(xpath, "//la/trigger/trigger%c/ch[%d]/text()", A_or_B, i + 1);
		channel = (unsigned char) read_long("Trigger Channel", xpath, -1);
		if (channel == -1)
		{
			continue;
		}

		/* Trigger Bedingung auswerten und Trigger in der Konfiguration aktualisieren */
		if (parse_la_trigger_condition(condition, &(trigger[channel])) != 0)
		{
			printf("Warnung: Ungültige Trigger Bedingung für Channel %d\n", channel);
		}
	}

	return 0;
}

/* Trigger auf Standardwerte setzen */
void set_default_la_trigger_values(struct la_config *cfg)
{
	int i;

	cfg->trigger_ignore = 1;

	/* Alle Trigger deaktivieren */
	for (i = 0; i < LA_NUM_CHANNELS; i++)
	{
		cfg->triggerA[i].low		= 0;
		cfg->triggerA[i].high		= 0;
		cfg->triggerA[i].falling_edge	= 0;
		cfg->triggerA[i].rising_edge	= 0;

		cfg->triggerB[i].low		= 0;
		cfg->triggerB[i].high		= 0;
		cfg->triggerB[i].falling_edge	= 0;
		cfg->triggerB[i].rising_edge	= 0;
	}

	/* Trigger Helfer 0 und 1 sind irgendwie sinnlos, daher werden die
	   Standardwerte so gesetzt, dass sie keine Bedeutung haben */
	cfg->trigger_helper0		= 0; /* 0 = A, 1 = NOT A, 2 = B, 3 = NOT B, 4 = A AND B, 5 = A OR B */
	cfg->trigger_helper1		= 2; /* 0 = A, 1 = NOT A, 2 = B, 3 = NOT B, 4 = A AND B, 5 = A OR B */
	cfg->trigger_selection		= 3; /* 0 = H1, 1 = H2, 2 = H1 AND H2, 3 = H1 OR H2 */
	cfg->trigger_start_trigger	= 0; /* 0 = wenn gültig, 1 = n-mal gültig */
	cfg->trigger_counter		= 0;
}

static char *trigger_selection_s[] = {"A", "B", "A AND B", "A OR B"};

/*
XML Datei mit den Konfigurationsdaten einlesen
*/
int read_la_config(struct la_config *cfg, const char *filename)
{
	int i;
	char buf[100];

	/* XML öffnen */
	if (xml_open(filename) != 0)
	{
		return -1;
	}

	/* Abtastrate einlesen */
	cfg->sample_rate = read_long("Abtastrate", "//la/samplerate/text()", 600000000);
	if (!(cfg->sample_rate > 0 && cfg->sample_rate <= 150000000) && cfg->sample_rate != 600000000)
	{
		printf("Warnung: Abtastrate ungültig.\n");
		cfg->sample_rate = 600000000;
	}

	/* Pretrigger einlesen */
	cfg->pretrigger = read_long("Pre-Trigger", "//la/pretrigger/text()", 1);
	if (cfg->pretrigger < 1 || cfg->pretrigger > 7)
	{
		printf("Warnung: Wert für Pre-Trigger ungültig.\n");
		cfg->pretrigger = 1;
	}

	/* Ignoretrigger einlesen */
	/*
	cfg->trigger_ignore = read_long("Ignore-Trigger", "//la/ignoretrigger/text()", 1);
	if (cfg->trigger_ignore != 0 && cfg->trigger_ignore != 1)
	{
		printf("Warnung: Wert für Ignore-Trigger ungültig.\n");
		cfg->trigger_ignore = 1;
	}
	*/

	/* Signale einlesen */
	cfg->num_signals = max(xml_get_element_count("//la/signals/signal"), 0);
	cfg->signals = (struct la_signal *)malloc(cfg->num_signals * sizeof(struct la_signal));

	for (i = 0; i < cfg->num_signals; i++)
	{
		read_la_signal(&(cfg->signals[i]), i);
	}

	/* Trigger A und B einlesen */
	set_default_la_trigger_values(cfg);
	if (xml_get_element_count("//la/trigger") > 0)
	{
		cfg->trigger_ignore = 0;
	}

	read_la_trigger(cfg->triggerA, 'A');
	read_la_trigger(cfg->triggerB, 'B');

	/* Andere Trigger Bedingungen einlesen */
	if (xml_get_string("//la/trigger/@when", buf, 100) == 0)
	{
		if (strcmp(buf, trigger_selection_s[0]) == 0)
		{
			cfg->trigger_selection = 0;
		}
		else if (strcmp(buf, trigger_selection_s[1]) == 0)
		{
			cfg->trigger_selection = 1;
		}
		else if (strcmp(buf, trigger_selection_s[2]) == 0)
		{
			cfg->trigger_selection = 2;
		}
		else if (strcmp(buf, trigger_selection_s[3]) == 0)
		{
			cfg->trigger_selection = 3;
		}
		else
		{
			printf("Warnung: Ungültige Trigger Bedingung %s.\n", buf);
		}
	}

	/* Trigger Zähler verfügbar? */
	if (xml_get_string("//la/trigger/@counter", buf, 100) == 0)
	{
		cfg->trigger_start_trigger = 1;
		cfg->trigger_counter = read_long("Trigger Counter", "//la/trigger/@counter", 1);
	}

	xml_close();

	return 0;
}

/* Speicher in der Konfiguration wieder freigeben */
void free_config(struct la_config *cfg)
{
	int i;

	for (i = 0; i < cfg->num_signals; i++)
	{
		if (cfg->signals[i].name != NULL)
		{
			free(cfg->signals[i].name);
			cfg->signals[i].name = NULL;
		}
		if (cfg->signals[i].channels != NULL)
		{
			free(cfg->signals[i].channels);
			cfg->signals[i].channels = NULL;
		}
	}

	if (cfg->signals != NULL)
	{
		free(cfg->signals);
		cfg->signals = NULL;
	}
}

/* Array der verfügbaren Zeit-Skalierungen für GTKWave */
/*                          s, ms, us, ns, ps, fs */
static int timescales[] = {0, -3, -6, -9, -12, -15};

/* Array der verfügbaren Zeit-Skalierungen als String */
static char *timescales_s[] = {"s", "ms", "us", "ns", "ps", "fs"};

/* Vergleich von Fließkommazahlen mit großzügigem epsilon */
#define float_cmp(a, b) (fabs(a - b) < 0.01)

/* Passende Zeitskalierung für die gegebene Abtastrate bestimmen.
   Bei einer Periodendauer von beispielsweise 20 ns wird die Skalierung auf 1 ns festgelegt.
   Bei einer Periodendauer von 20,666 ns wird die Skalierung auf 1 ps festgelegt um die Genauigkeit in der Waveform zu erhöhen.
   Der Rückgabewert ist der Index im timescales Array.
   Der Ausgabeparameter delta erhält die Schrittweite für GTKWave. */
int get_timescale_index(unsigned long sample_rate, int *delta)
{
	int index;
	float rate, period;

	/* Abtastrate so lange durch 1000 teilen, bis sie kleiner als 1 ist.
           In jedem Schritt den Index um 1 erhöhen (entspricht der nächstkleineren Zeiteinheit) */
	index = 0;
	rate = sample_rate;
	while (rate > 1)
	{
		index++;
		rate /= 1000;
	}

	/* Periodendauer berechnen */
	period = 1.0 / rate;
	/* Wenn die Periode noch Nachkommastellen hat, die nächstkleinere Zeiteinheit wählen */
	if (!float_cmp(floor(period), period))
	{
		index++;
		period *= 1000;
	}

	/* Index darf nicht größer als 5 sein, sonst liegt er außerhalb des timescales Arrays */
	if (index > 5)
	{
		index = 5;
	}

	if (delta != NULL)
	{
		*delta = (int) period;
	}

	return index;
}

/*
lxt Datei mit den Messdaten erzeugen
*/
int write_lxt(const char *filename, struct la_config *cfg, unsigned long long *data, unsigned long length)
{
	int i, j, k;
	int timescale_index, timescale, delta;
	unsigned long base_addr;
	struct lt_trace *lt;
	struct lt_symbol **symbols;

	/* Skalierung für die Zeitachse und entsprechende Schrittweite pro Sample berechnen */
	timescale_index = get_timescale_index(cfg->sample_rate, &delta);
	timescale = timescales[timescale_index];

	/* LXT Datei öffnen */
	lt = lt_init(filename);
	if (lt == NULL)
	{
		return -1;
	}

	/* Zeitskalierung setzen */
	lt_set_initial_value(lt, 'U');
	lt_set_timescale(lt, -timescale);
	lt_set_time(lt, 0);

	/* Signale (Symbole) erzeugen */
	symbols = (struct lt_symbol **)malloc(cfg->num_signals * sizeof(struct lt_symbol *));

	for (i = 0; i < cfg->num_signals; i++)
	{
		symbols[i] = lt_symbol_add(lt, cfg->signals[i].name, 0, cfg->signals[i].num_channels - 1, 0, LT_SYM_F_BITS);
	}

	base_addr = la_get_addr_cnt();
	la_test_error(1);

	/* Für alle Datensätze der Messdaten... */
	for (i = 0; i < length; i++)
	{
		int ram_index, ram_addr;
		unsigned long long sample;

		/* Die 4 internen RAMs werden abwechselnd mit aufsteigenden Adressen ausgelesen.
		   Der erste Messwert steht im zweiten RAM (Index=1) an Adresse 0.
		   Der letzte gültige Messwert steht im ersten RAM (Index=0) an Adresse 0. */
		ram_index = ((i + 1) & 3) * 1024;
		ram_addr = ((base_addr + ((i + 1) & 0xffc)) & 0xfff) >> 2;

		sample = data[ram_index + ram_addr];

		/* Für alle erzeugten Signale... */
		for (j = 0; j < cfg->num_signals; j++)
		{
			int value = 0;

			/* Für alle Channels dieses Signals */
			for (k = 0; k < cfg->signals[j].num_channels; k++)
			{
				value |= ((sample >> cfg->signals[j].channels[k]) & 1) << k;
			}

			lt_emit_value_int(lt, symbols[j], 0, value);
		}

		lt_inc_time_by_delta(lt, delta);
	}

	lt_close(lt);

	free(symbols);

	return 0;
}

/*
.sav Datei für gtkwave erzeugen
*/
int write_sav(const char *filename, struct la_config *cfg)
{
	int i;
	int timescale_index, timescale, delta;
	FILE *fp;

	fp = fopen(filename, "w");
	if (fp == NULL)
	{
		return -1;
	}

	/* Marker Position für Pre-Trigger berechnen */
	timescale_index = get_timescale_index(cfg->sample_rate, &delta);
	timescale = timescales[timescale_index];
	fprintf(fp, "*0 0 %f\n", 512.0 * delta * cfg->pretrigger);
	
	/* Alle Signale in die Anzeigeliste aufnehmen */
	for (i = 0; i < cfg->num_signals; i++)
	{
		fputs(cfg->signals[i].name, fp);
		/* Falls Signal ein Bus ist, die Breitenangabe dazuschreiben */
		if (cfg->signals[i].num_channels > 1)
		{
			fprintf(fp, "[%d:0]", cfg->signals[i].num_channels - 1);
		}
		fputs("\n", fp);
	}

	fclose(fp);

	return 0;
}

void print_headline(char *title, int width, char symbol)
{
	char *line;

	line = (char *)malloc(width - 1);
	memset(line, symbol, width - 2);
	line[width - 2] = '\0';

	printf("+%s+\n| %-*s |\n+%s+\n\n", line, width - 4, title, line);

	free(line);
}

void printBinaryValue(unsigned long long value)
{
	int i;
	for (i = 35; i >= 0; i--)
	{
		printf("%c", (value & (1 << i)) ? '1' : '0');
		if (i % 8 == 0)
		{
			printf(" ");
		}
	}
}

void printStatusBar(float maxProgress, int busyStatus)
{
	int i;
	int la_status, progress, busy;

	printf("[                                                  ]\r[");

	/* Lokale Variablen initialisieren */
	la_status = la_get_status(NULL);
	la_test_error(1);
	i = 0;
	busy = 1;

	/* Solange sich der LA Status nicht geändert hat... */
	while (busy)
	{
		/* Frage aktuellen Status ab */
		la_status = la_get_status(&progress);
		la_test_error(1);

		/* Wenn sich der Status geändert hat, Fortschritt auf 100% setzen und Schleife verlassen */
		if (la_status != busyStatus)
		{
			progress = 100;
			busy = 0;
		}
		/* Aktuellen Fortschritt berechnen */
		else
		{
			progress = (int) ((100 / maxProgress) * progress);
		}
		
		/* Fortschrittsbalken zeichnen */
		while (i < progress)
		{
			printf("|");
			i += 2;
		}
		/* Ausgabepuffer der Standardausgabe leeren */
		fflush(stdout);
	}

	printf("]\n\n");
}

int main(int argc, char **argv)
{
	int i;
	char filenames[3][60];
	unsigned long long data[4096];
	struct la_config cfg;

	print_headline("USB-LOGI-500 Logik Analysator v1.0", 60, '=');

	if (argc < 2)
	{
		printf("Fehler: Keinen Projektnamen angegeben.\n\n");
		printf("Aufruf: ./la_sample <Projektname>\n");
		return 1;
	}

	/* Speicherüberlauf verhindern */
	if (strlen(argv[1]) > 50)
	{
		argv[1][50] = '\0';
	}

	/* Dateinamen zusammenbauen für xml, lxt und sav Datei */
	sprintf(filenames[0], "%s.xml", argv[1]);
	sprintf(filenames[1], "%s.lxt", argv[1]);
	sprintf(filenames[2], "%s.sav", argv[1]);

	/* Konfiguration aus XML Datei lesen */
	print_headline("(1) Lese Projektdatei...", 60, '-');
	printf("Projektdatei: %s\n\n", filenames[0]);
	if (read_la_config(&cfg, filenames[0]) != 0)
	{
		printf("Fehler: Projektdatei konnte nicht gelesen werden.\n");
		return 1;
	}
	else
	{
		int timescale_index, period;
		char buf[32];
		unsigned long long triggerA[4], triggerB[4];

		timescale_index = get_timescale_index(cfg.sample_rate, &period);
		sprintf(buf, "%d %s", period, timescales_s[timescale_index]);

		la_build_trigger_reg_values(cfg.triggerA, triggerA);
		la_build_trigger_reg_values(cfg.triggerB, triggerB);

		printf("+----------------------------------------------------------+\n");
		printf("| Messung mit folgenden Einstellungen                      |\n");
		printf("+----------------------------------------------------------+\n");
		printf("| Abtastrate (Samples/s) | %-31ld |\n", cfg.sample_rate);
		printf("| Abtastperiode          | %-31s |\n", buf);
		printf("| Trigger ignorieren     | %-31s |\n", cfg.trigger_ignore ? "ja" : "nein");
		printf("| Pretrigger             | %1d/8                             |\n", cfg.pretrigger);
		printf("| Trigger Auswahl        | %-31s |\n", trigger_selection_s[cfg.trigger_selection]);
		printf("| Trigger Aktivierung    | %-32s |\n", (cfg.trigger_start_trigger == 0) ? "wenn gültig" : "wenn n-mal gültig");
		if (cfg.trigger_start_trigger == 1)
		{
			printf("| Trigger Zähler         | %-31ld |\n", cfg.trigger_counter);
		}
		printf("+----------------------------------------------------------+\n");
		printf("| Trigger A   CH 35 32 31    24 23    16 15     8 7      0 |\n");
		printf("+----------------------------------------------------------+\n");
		printf("| LOW          |  "); printBinaryValue(triggerA[0]); printf("|\n");
		printf("| HIGH         |  "); printBinaryValue(triggerA[1]); printf("|\n");
		printf("| RISING_EDGE  |  "); printBinaryValue(triggerA[2]); printf("|\n");
		printf("| FALLING_EDGE |  "); printBinaryValue(triggerA[3]); printf("|\n");
		printf("+----------------------------------------------------------+\n");
		printf("| Trigger B   CH 35 32 31    24 23    16 15     8 7      0 |\n");
		printf("+----------------------------------------------------------+\n");
		printf("| LOW          |  "); printBinaryValue(triggerB[0]); printf("|\n");
		printf("| HIGH         |  "); printBinaryValue(triggerB[1]); printf("|\n");
		printf("| RISING_EDGE  |  "); printBinaryValue(triggerB[2]); printf("|\n");
		printf("| FALLING_EDGE |  "); printBinaryValue(triggerB[3]); printf("|\n");
		printf("+----------------------------------------------------------+\n");
		printf("\n");

	}

	/* Zum USB-LOGI verbinden und konfigurieren */
	print_headline("(2) Verbinde zum USB-LOGI...", 60, '-');

	if (la_connect() != 0)
	{
		printf("Fehler: Es konnte keine Verbindung aufgebaut werden.\n");
		la_test_error(0);
		return 2;
	}

	printf("Verbindung hergestellt.\n\n");
	
	if (la_configure(&cfg) != 0)
	{
		printf("Fehler: Gerätekonfiguration fehlgeschlagen.\n");
		la_test_error(0);
		return 2;
	}
	la_stop_sampling();

	printf("Geräte-Konfiguration erfolgreich.\n\n");
	

	/* Messung starten */
	print_headline("(3) Starte Messung...", 60, '-');

	if (la_start_sampling() != 0)
	{
		printf("Fehler: Verbindung zum USB-LOGI verloren.\n");
		la_test_error(0);
		return 3;
	}

	printf("Fülle Pre-Trigger...\n\n");
	printStatusBar(12.5f * cfg.pretrigger, LA_STATUS_FILLING_PRETRIGGER);

	if (cfg.trigger_ignore == 0)
	{
		unsigned long last_trigger_cnt;

		printf("Warte auf Trigger...\n\n");

		last_trigger_cnt = la_get_trigger_cnt();
		while (la_get_status(NULL) == LA_STATUS_WAITING_FOR_TRIGGER)
		{
			if (cfg.trigger_start_trigger == 1)
			{
				if (la_get_trigger_cnt() != last_trigger_cnt)
				{
					last_trigger_cnt = la_get_trigger_cnt();
					printf("Trigger Bedingung wurde bisher %ld mal erfüllt.\r", last_trigger_cnt);
					fflush(stdout);
				}
			}
			la_test_error(1);
		}
		if (cfg.trigger_start_trigger == 1)
		{
			printf("\n\n");
		}
	}

	printf("Aufzeichnung läuft...\n\n");
	printStatusBar(100 - (12.5f * cfg.pretrigger), LA_STATUS_SAMPLING);

	/* Messdaten auslesen */
	if (la_get_sampling_data(data, 4096) != 0)
	{
		printf("Fehler: Verbindung zum USB-LOGI verloren.\n");
		la_test_error(0);
		return 3;
	}

	/* LXT und SAV Datei für GTKWave erzeugen */
	print_headline("(4) Erzeuge Waveform Dateien...", 60, '-');

	if (write_lxt(filenames[1], &cfg, data, 4096) != 0)
	{
		printf("Fehler: %s konnte nicht geschrieben werden.\n", filenames[1]);
		return 4;
	}

	printf("LXT Datei %s erzeugt.\n\n", filenames[1]);
	
	if (write_sav(filenames[2], &cfg) != 0)
	{
		printf("Fehler: %s konnte nicht geschrieben werden.\n", filenames[1]);
		return 4;
	}

	printf("SAV Datei %s erzeugt.\n\n", filenames[2]);

	/* Verbindung zum USB-LOGI trennen */
	la_disconnect();

	/* Konfigurationsspricher freigeben */
	free_config(&cfg);

	/* GTKWave starten */
	print_headline("(5) Starte GTKWave...", 60, '-');
	execlp("gtkwave", "gtkwave", filenames[1], filenames[2], NULL);

	return 0;
}

